/************************************************************************
 *
 *			ob|rdqud
 *
 *				: f[^t@C
 *		t@C	: dbf.c
 *		쐬		: s.aizawa
 *
 ************************************************************************/

#include "pcsos.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <io.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "pcsdef.h"
#include "pcsdb.h"
#include "pcsdbsub.h"
#include "pcscom.h"

static	SHORT	_DBFOpen(CHAR *, PDBF *);
static	SHORT	_DBFCreate(CHAR *, PDBF_I, SHORT, PDBF *);
static	SHORT	_DBFClose(PDBF);

static	PDBF	pTopDBF;
CRITICALSECTION	csDBF;		/* DBF N`JZNV */

/*=======================================================================
 |
 |		f[^t@CI[v
 |
 |	SHORT	DBFOpen(fileName, &dpp)
 |
 |		CHAR	*fileName;	f[^t@C
 |		PDBF	*dpp;		cae|C^
 |
 |		Ԓl			G[R[h
 |
 =======================================================================*/
SHORT	DBFOpen(CHAR *fileName, PDBF *dpp)
{
	SHORT	err;

	ENTER_CRITICAL_SECTION(&csDBF);
	err = _DBFOpen(fileName, dpp);
	LEAVE_CRITICAL_SECTION(&csDBF);

	return err;
}

static	SHORT	_DBFOpen(CHAR *fileName, PDBF *dpp)
{
	PDBF	dp;
	PDBF_H	dhp;
	PFIELD	flp;
	SHORT	i;
	SHORT	offset;
	CHAR	pathName[256];
	LONG	nRec;
	CHAR	buf[2];
#if defined (OS_WNT)
	DWORD	numBytes;
#endif
	BOOL	fBinaryField;
	CHAR	binFileName[256];
	CHAR	*p;
	SHORT	err;

	/* tpX擾 */
	_fullpath(pathName, fileName, sizeof(pathName));
	StringToUpper(pathName);
	for (dp = pTopDBF; dp != NULL; dp = dp->chain) {
		if (strcmp(pathName, dp->fileName) == 0) {
#if defined (MULTI_OPEN)
			if ((dp->fh = __OPEN(pathName)) == -1)
				return ERROR_DB_FOPEN;
#endif
			*dpp = dp;
			return 0;
		}
	}

	/* f[^t@CǗGAm */
	if ((dp = malloc(sizeof(DBF))) == 0)
		return ERROR_DB_MEMORY;

	memset(dp, 0, sizeof(DBF));
	INIT_CRITICAL_SECTION(&dp->csDBF);

	/* f[^t@CI[v */
	if ((dp->fh = __OPEN(pathName)) == -1) {
		_DBFClose(dp);
		return ERROR_DB_NOFILE;
	}

	strcpy(dp->fileName, pathName);	/* t@CZ[u */

	/* wb_GAm */
	if ((dhp = malloc(sizeof(DBF_H))) == 0) {
		_DBFClose(dp);
		return ERROR_DB_MEMORY;
	}
	dp->dhp = dhp;

	/* wb_ǂݍ */
	if (__READ(dp->fh, (CHAR *)dhp, sizeof(DBF_H)) != sizeof(DBF_H)) {
		_DBFClose(dp);
		return ERROR_DB_READ;
	}

	/* tB[hGAm */
	dp->nField = (dhp->lHeader - 33) / 32;
	if ((dp->flp = malloc(sizeof(FIELD) * dp->nField)) == 0) {
		_DBFClose(dp);
		return ERROR_DB_MEMORY;
	}

	/* tB[hǂݍ */
	if (__READ(dp->fh, (CHAR *)dp->flp, sizeof(FIELD) * dp->nField) !=
					(int)(sizeof(FIELD) * dp->nField)) {
		_DBFClose(dp);
		return ERROR_DB_READ;
	}

	/* tB[h|C^Zbg */
	flp = dp->flp;
	offset = 0;
	fBinaryField = FALSE;
	for (i = 0; i < dp->nField; i++) {
		if (dhp->extField != EXT_FIELD)
			flp->lField2 = flp->lField;
		flp->offset = offset;
		offset += flp->lField2;

		if (flp->type == 'B')
			fBinaryField = TRUE;
		else
			dp->lRec += flp->lField2;

		flp++;
	}

	/* oCit@CI[v */
	if (fBinaryField) {
		strcpy(binFileName, fileName);
		if ((p = strrchr(binFileName, '.')) == NULL)
			p = binFileName + strlen(binFileName);
		strcpy(p, ".BIN");
		if ((err = BINOpen(binFileName, &dp->bfp)) != 0) {
			if (err == ERROR_DB_NOFILE && (err = BINCreate(binFileName, &dp->bfp)) != 0) {
				_DBFClose(dp);
				return err;
			}
		}
	}

	/* R[h */
	nRec = (__LSEEK(dp->fh, 0L, SEEK_END) - dhp->lHeader) / dhp->lRec;
	if (nRec < 0)
		nRec = 0;
	if (nRec > dhp->nRec) {
		__LSEEK(dp->fh, dhp->lHeader + dhp->nRec * dhp->lRec, SEEK_SET);
		if (__READ(dp->fh, buf, 1) == 1) {
			if (buf[0] == 0x1a)
				nRec = dhp->nRec;
		}
	}
	if (nRec != dhp->nRec) {
		dhp->nRec = nRec;
		dp->updateFlag = TRUE;	/* t@CXVtOZbg */
		dp->adjustFlag = TRUE;
	}

	/* `FC */
	dp->chain = pTopDBF;
	pTopDBF = dp;

	*dpp = dp;

	return 0;
}

/*=======================================================================
 |
 |		f[^t@C쐬
 |
 |	SHORT	DBFCreate(fileName, dip, nField, dpp)
 |
 |		CHAR	*fileName;	f[^t@C
 |		PDBF_I	dip;		f[^t@C
 |		SHORT	nField;		tB[h
 |		PDBF	*dpp;		cae|C^
 |
 |		Ԓl			G[R[h
 |
 =======================================================================*/
SHORT	DBFCreate(CHAR *fileName, PDBF_I dip, SHORT nField, PDBF *dpp)
{
	SHORT	err;

	ENTER_CRITICAL_SECTION(&csDBF);
	err = _DBFCreate(fileName, dip, nField, dpp);
	LEAVE_CRITICAL_SECTION(&csDBF);

	return err;
}

static	SHORT	_DBFCreate(CHAR *fileName, PDBF_I dip, SHORT nField, PDBF *dpp)
{
	PDBF	dp;
	PDBF_H	dhp;
	PFIELD	flp;
	SHORT	i;
	SHORT	offset;
	BOOL	fExtField;
	CHAR	pathName[256];
	static	CHAR	cr_eof[2] = {0x0d, 0x1a};
#if defined (OS_WNT) && !defined (_SERVER)
	DWORD	numBytes;
#endif
	BOOL	fBinaryField;
	CHAR	binFileName[256];
	CHAR	*p;
	SHORT	err;

	/* tpX擾 */
	_fullpath(pathName, fileName, sizeof(pathName));
	StringToUpper(pathName);

	for (dp = pTopDBF; dp != NULL; dp = dp->chain) {
		if (strcmp(pathName, dp->fileName) == 0) {
			return ERROR_DB_ARDYOPEN;
		}
	}

	/* f[^t@CǗGAm */
	if ((dp = malloc(sizeof(DBF))) == 0)
		return ERROR_DB_MEMORY;

	memset(dp, 0, sizeof(DBF));
	INIT_CRITICAL_SECTION(&dp->csDBF);

	/* f[^t@CI[v */
	if ((dp->fh = __CREAT(pathName)) == -1) {
		_DBFClose(dp);
		return ERROR_DB_FOPEN;
	}

	strcpy(dp->fileName, pathName);	/* t@CZ[u */

	/* wb_GAm */
	if ((dhp = malloc(sizeof(DBF_H))) == 0) {
		_DBFClose(dp);
		return ERROR_DB_MEMORY;
	}
	memset(dhp, 0, sizeof(DBF_H));
	dp->dhp = dhp;

	/* wb_ݒ */
	dhp->flag = 0x03;
	dhp->lHeader = (nField + 1) * 32 + 1;

	/* tB[hGAm */
	dp->nField = nField;
	if ((dp->flp = malloc(sizeof(FIELD) * nField)) == 0) {
		_DBFClose(dp);
		return ERROR_DB_MEMORY;
	}
	memset(dp->flp, 0, sizeof(FIELD) * nField);

	fExtField = FALSE;
	for (i = 0; i < nField; i++) {
		if (dip[i].lField > 255) {
			fExtField = TRUE;
			dhp->extField = EXT_FIELD;
			dhp->flag = 0x04;
			break;
		}
	}

	/* tB[hݒ */
	fBinaryField = FALSE;
	for (i = 0, flp = dp->flp, dhp->lRec = 1; i < nField; i++, flp++) {
		strcpy(flp->name, dip->name);
		StringToUpper(flp->name);
		flp->type = toupper(dip->type);

		/* tB[hʃ`FbN */
		switch (flp->type) {
		case 'B':
			fBinaryField = TRUE;
			flp->lField2 = 4;
			break;
		case 'C':
		case 'D':
		case 'I':
		case 'L':
		case 'N':
			if (fBinaryField != 0) {
				_DBFClose(dp);
				return ERROR_DB_FIELD;
			}

			flp->lField2 = dip->lField;
			dp->lRec += flp->lField2;
			break;
		default:
			_DBFClose(dp);
			return ERROR_DB_FIELD;
		}

		if (flp->type == 'N')
			flp->lDec = (CHAR)dip->lDec;
		else
			flp->lDec = (CHAR)0;

		flp->lField = (CHAR)flp->lField2;
		dhp->lRec += flp->lField2;
		dip++;
	}

	/* oCit@C쐬 */
	if (fBinaryField) {
		strcpy(binFileName, fileName);
		if ((p = strrchr(binFileName, '.')) == NULL)
			p = binFileName + strlen(binFileName);
		strcpy(p, ".BIN");
		if ((err = BINCreate(binFileName, &dp->bfp)) != 0) {
			_DBFClose(dp);
			return err;
		}
	}

	/* wb_񏑂 */
	if (__WRITE(pathName, dp->fh, (CHAR *)dp->dhp, sizeof(DBF_H)) !=
							sizeof(DBF_H)) {
		_DBFClose(dp);
		return ERROR_DB_WRITE;
	}

	/* tB[h񏑂 */
	if (__WRITE(pathName, dp->fh, (CHAR *)dp->flp,
				sizeof(FIELD) * dp->nField) !=
				(int)(sizeof(FIELD) * dp->nField)) {
		_DBFClose(dp);
		return ERROR_DB_WRITE;
	}

	/* ؂蕶dne */
	if (__WRITE(pathName, dp->fh, cr_eof, 2) != 2) {
		_DBFClose(dp);
		return ERROR_DB_WRITE;
	}

	/* tB[h|C^Zbg */
	flp = dp->flp;
	offset = 0;
	for (i = 0; i < nField; i++) {
		flp->offset = offset;
		offset += flp->lField2;
		flp++;
	}

	dp->updateFlag = 1;
	dp->updateCount++;

	/* `FC */
	dp->chain = pTopDBF;
	pTopDBF = dp;

	*dpp = dp;

	return 0;
}

/*=======================================================================
 |
 |		R[hǂݍ݁iR[hŚj
 |
 |	SHORT	DBFRead(dp, rno, bufp)
 |
 |		PDBF	dp;		cae|C^
 |		LONG	rno;		R[hԍ
 |		CHAR	*bufp;		obt@|C^
 |
 |		Ԓl			G[R[h
 |
 =======================================================================*/
SHORT	DBFRead(PDBF dp, LONG rno, CHAR *bufp)
{
	PDBF_H	dhp;
#if defined (OS_WNT)
	DWORD	numBytes;
#endif

	dhp = dp->dhp;

	/* R[hԍ͈͊OȂG[ */
	if (rno <= 0 || rno > dhp->nRec)
		return ERROR_DB_RECNO;

	/* ǂݍ */
	__LSEEK(dp->fh, dhp->lHeader + dhp->lRec*(rno-1), SEEK_SET);
	if (__READ(dp->fh, bufp, dhp->lRec) != (int)dhp->lRec)
		return ERROR_DB_READ;

	if (dhp->scramble != 0)
		Scramble2(bufp + 1, dp->lRec, dhp->scramble);

	return 0;
}

/*=======================================================================
 |
 |		R[hǂݍ݁i폜}[NCoCitB[hj
 |
 |	SHORT	DBFRead2(dp, rno, bufp)
 |
 |		PDBF	dp;		cae|C^
 |		LONG	rno;		R[hԍ
 |		CHAR	*bufp;		obt@|C^
 |
 |		Ԓl			G[R[h
 |
 =======================================================================*/
SHORT	DBFRead2(PDBF dp, LONG rno, CHAR *bufp)
{
	PDBF_H	dhp;
#if defined (OS_WNT)
	DWORD	numBytes;
#endif

	dhp = dp->dhp;

	/* R[hԍ͈͊OȂG[ */
	if (rno <= 0 || rno > dhp->nRec)
		return ERROR_DB_RECNO;

	/* ǂݍ */
	__LSEEK(dp->fh, dhp->lHeader + dhp->lRec*(rno-1) + 1, SEEK_SET);
	if (__READ(dp->fh, bufp, dp->lRec) != (int)dp->lRec)
		return ERROR_DB_READ;

	if (dhp->scramble != 0)
		Scramble2(bufp, dp->lRec, dhp->scramble);

	return 0;
}

/*=======================================================================
 |
 |		R[h݁iR[hŚj
 |
 |	SHORT	DBFWrite(dp, rno, bufp)
 |
 |		PDBF	dp;		cae|C^
 |		LONG	rno;		R[hԍ
 |		CHAR	*bufp;		obt@|C^
 |
 |		Ԓl			G[R[h
 |
 =======================================================================*/
SHORT	DBFWrite(PDBF dp, LONG rno, CHAR *bufp)
{
	PDBF_H	dhp;
	int	rc;
#if defined (OS_WNT) && !defined (_SERVER)
	DWORD	numBytes;
#endif

	dhp = dp->dhp;

	/* R[hԍ͈͊OȂG[ */
	if (rno <= 0 || rno > dhp->nRec)
		return ERROR_DB_RECNO;

	if (dhp->scramble != 0)
		Scramble1(bufp + 1, dp->lRec, dhp->scramble);

	/*  */
	rc = __WRITE2(dp->fileName, dp->fh, bufp, dhp->lRec,
				dhp->lHeader + dhp->lRec*(rno-1));

	if (dhp->scramble != 0)
		Scramble2(bufp + 1, dp->lRec, dhp->scramble);

	if (rc != (int)dhp->lRec)
		return ERROR_DB_WRITE;

	dp->updateFlag = 1;		/* t@CXVtOZbg */
	dp->updateCount++;

	return 0;
}

/*=======================================================================
 |
 |		R[h݁i폜}[NCoCitB[hj
 |
 |	SHORT	DBFWrite2(dp, rno, bufp)
 |
 |		PDBF	dp;		cae|C^
 |		LONG	rno;		R[hԍ
 |		CHAR	*bufp;		obt@|C^
 |
 |		Ԓl			G[R[h
 |
 =======================================================================*/
SHORT	DBFWrite2(PDBF dp, LONG rno, CHAR *bufp)
{
	PDBF_H	dhp;
	int	rc;
#if defined (OS_WNT) && !defined (_SERVER)
	DWORD	numBytes;
#endif

	dhp = dp->dhp;

	/* R[hԍ͈͊OȂG[ */
	if (rno <= 0 || rno > dhp->nRec)
		return ERROR_DB_RECNO;

	if (dhp->scramble != 0)
		Scramble1(bufp, dp->lRec, dhp->scramble);

	/*  */
	rc = __WRITE2(dp->fileName, dp->fh, bufp, dp->lRec,
					dhp->lHeader + dhp->lRec*(rno-1) + 1);

	if (dhp->scramble != 0)
		Scramble2(bufp, dp->lRec, dhp->scramble);

	if (rc != (int)dp->lRec)
		return ERROR_DB_WRITE;

	dp->updateFlag = 1;		/* t@CXVtOZbg */
	dp->updateCount++;

	return 0;
}

/*=======================================================================
 |
 |		f[^t@CN[Y
 |
 |	SHORT	DBFClose(dp)
 |
 |		PDBF	dp;		cae|C^
 |
 |		Ԓl			G[R[h
 |
 =======================================================================*/
SHORT	DBFClose(PDBF dp)
{
	SHORT	err;

	ENTER_CRITICAL_SECTION(&csDBF);
	err = _DBFClose(dp);
	LEAVE_CRITICAL_SECTION(&csDBF);

	return err;
}

static	SHORT	_DBFClose(PDBF dp)
{
	PDBF	*ppDBF;

	if (dp->fh != -1) {
		DBFFlush(dp, FALSE);	/* t@C */
		__CLOSE(dp->fh);	/* t@CN[Y */
	}

	if (dp->bfp != NULL)
		BINClose(dp->bfp);	/* oCit@CN[Y */

	if (dp->dhp != NULL)
		free(dp->dhp);		/* wb_GA */

	if (dp->flp != NULL)
		free(dp->flp);		/* tB[hGA */

	/* `FC؂藣 */
	for (ppDBF = &pTopDBF; *ppDBF != 0; ppDBF = &(*ppDBF)->chain) {
		if ((*ppDBF) == dp) {
			*ppDBF = (*ppDBF)->chain;
			break;
		}
	}

	DELETE_CRITICAL_SECTION(&dp->csDBF);

	free(dp);		/* f[^t@CǗGA */

	return 0;
}

/*=======================================================================
 |
 |		f[^t@C|o
 |
 |	SHORT	DBFFlush(dp, bReopen)
 |
 |		PDBF	dp;		cae|C^
 |		BOOL	bReopen;	ăI[vtO
 |
 |		Ԓl			G[R[h
 |
 =======================================================================*/
SHORT	DBFFlush(PDBF dp, BOOL bReopen)
{
	static	CHAR	eof[1] = {0x1a};
	PDBF_H	dhp;
	LONG	ltime;
	struct	tm	*tmp;
#if defined (OS_WNT) && !defined (_SERVER)
	DWORD	numBytes;
#endif

	if (dp->updateFlag != 0) {
		dhp = dp->dhp;

		/* ŏIXV */
		time(&ltime);
		tmp = localtime(&ltime);
		dhp->date[0] = (CHAR)tmp->tm_year;
		dhp->date[1] = (CHAR)(tmp->tm_mon + 1);
		dhp->date[2] = (CHAR)tmp->tm_mday;

		/* wb_ */
		if (__WRITE2(dp->fileName, dp->fh, (CHAR *)dp->dhp,
					sizeof(DBF_H), 0) != sizeof(DBF_H))
			return ERROR_DB_WRITE;

		/* dne */
		if (__WRITE2(dp->fileName, dp->fh, eof, 1,
				dhp->lHeader + dhp->lRec * dhp->nRec) != 1)
			return ERROR_DB_WRITE;

		dp->updateFlag = 0;
	}

	if (bReopen) {
		__CLOSE(dp->fh);
		dp->fh = __OPEN(dp->fileName);
	} else
		__FLUSH(dp->fh);

	if (dp->bfp != NULL)
		BINFlush(dp->bfp, bReopen);

	return 0;
}

/*=======================================================================
 |
 |		폜}[NZbg
 |
 |	SHORT	DBFDelete(dp, rno)
 |
 |		PDBF	dp;		cae|C^
 |		LONG	rno;		R[hԍ
 |
 |		Ԓl			G[R[h
 |
 =======================================================================*/
SHORT	DBFDelete(PDBF dp, LONG rno)
{
	PDBF_H	dhp;
	CHAR	mark[1];
#if defined (OS_WNT) && !defined (_SERVER)
	DWORD	numBytes;
#endif

	dhp = dp->dhp;

	/* R[hԍ͈͊OȂG[ */
	if (rno <= 0 || rno > dhp->nRec)
		return ERROR_DB_RECNO;

	/* 폜tOZbg */
	mark[0] = '*';
	if (__WRITE2(dp->fileName, dp->fh, mark, 1,
				dhp->lHeader + dhp->lRec*(rno-1)) != 1)
		return ERROR_DB_WRITE;

	dp->updateFlag = 1;		/* t@CXVtOZbg */
	dp->updateCount++;

	return 0;
}

/*=======================================================================
 |
 |		폜}[N
 |
 |	SHORT	DBFRecall(dp, rno)
 |
 |		PDBF	dp;		cae|C^
 |		LONG	rno;		R[hԍ
 |
 |		Ԓl			G[R[h
 |
 =======================================================================*/
SHORT	DBFRecall(PDBF dp, LONG rno)
{
	PDBF_H	dhp;
	CHAR	mark[1];
#if defined (OS_WNT) && !defined (_SERVER)
	DWORD	numBytes;
#endif

	dhp = dp->dhp;

	/* R[hԍ͈͊OȂG[ */
	if (rno <= 0 || rno > dhp->nRec)
		return ERROR_DB_RECNO;

	/* 폜tOZbg */
	mark[0] = ' ';
	if (__WRITE2(dp->fileName, dp->fh, mark, 1,
				dhp->lHeader + dhp->lRec*(rno-1)) != 1)
		return ERROR_DB_WRITE;

	dp->updateFlag = 1;		/* t@CXVtOZbg */
	dp->updateCount++;

	return 0;
}

/*=======================================================================
 |
 |		폜}[NĂ郌R[ht@C폜
 |
 |	SHORT	DBFPack(dp)
 |
 |		PDBF	dp;		cae|C^
 |
 |		Ԓl			G[R[h
 |
 =======================================================================*/
SHORT	DBFPack(PDBF dp)
{
	PDBF_H	dhp;
	PBIN	bfp;
	PDB	dbp2;
	PDBLOCK	dlp;
	PFIELD	flp;
	LONG	r_rno, w_rno;
	CHAR	*rbp;
	CHAR	*fdp;
	int	i;
	SHORT	err = 0;
	CHAR	binFileName[256];
	CHAR	binTempFileName[256];
	CHAR	dbfBackupFileName[256];
	PBIN	bfpTemp;
	CHAR	*p;
	LONG	bno;
	BOOL	bWriteBinary;
	BOOL	bRc;

	dhp = dp->dhp;
	bfp = dp->bfp;

	if ((rbp = malloc(dhp->lRec)) == NULL)
		return ERROR_DB_MEMORY;

	/* 폜}[NtĂ郌R[h폜 */
	for (r_rno = w_rno = 0; r_rno < dhp->nRec; ) {
		if ((err = DBFRead(dp, ++r_rno, rbp)) != 0)
			break;

		if (rbp[0] != '*') {
			if (r_rno != ++w_rno) {
				if ((err = DBFWrite(dp, w_rno, rbp)) != 0)
					break;

				/* JgR[hԍ␳ */
				ENTER_CRITICAL_SECTION(&csDBF);
				for (dbp2 = dp->dbp; dbp2 != NULL;
							dbp2 = dbp2->chain) {
					if (dbp2->rno == r_rno)
						dbp2->rno = w_rno;
				}
				LEAVE_CRITICAL_SECTION(&csDBF);

				/* bNR[hԍ␳ */
				for (dlp = dp->rdlp; dlp != NULL;
							dlp = dlp->chain) {
					if (dlp->rno == r_rno)
						dlp->rno = w_rno;
				}
				for (dlp = dp->wdlp; dlp != NULL;
							dlp = dlp->chain) {
					if (dlp->rno == r_rno)
						dlp->rno = w_rno;
				}
			}
		} else {
			/* oCif[^̍폜 */
			if (bfp != NULL) {
				for (i = 0, flp = dp->flp; i < dp->nField; i++, flp++) {
					if (flp->type == 'B') {
						fdp = &rbp[flp->offset + 1];
						BINDelete(bfp, *(LONG *)fdp);
					}
				}
			}
		}
	}

	if (err != 0) {
		free(rbp);
		return err;
	}

	/* t@CTCY𒲐 */
	__CHSIZE(dp->fileName, dp->fh, dhp->lHeader + dhp->lRec * w_rno);

	/* R[hZbg */
	dhp->nRec = w_rno;

	/* oCif[^̃pbN */
	if (bfp != NULL && bfp->bhp->freeBlock != 0) {
		// DBF obNAbvt@C쐬
		strcpy(dbfBackupFileName, dp->fileName);
		if ((p = strrchr(dbfBackupFileName, '.')) == NULL)
			p = dbfBackupFileName + strlen(dbfBackupFileName);
		strcpy(p, ".BAK");

		// DBF obNAbvRs[
		__CLOSE(dp->fh);
		bRc = CopyFile(dp->fileName, dbfBackupFileName, FALSE);
		dp->fh = __OPEN(dp->fileName);
		if (!bRc) {
			free(rbp);
			return ERROR_DB_TMPFILE;
		}

		// BIN e|t@C쐬
		strcpy(binTempFileName, dp->fileName);
		if ((p = strrchr(binTempFileName, '.')) == NULL)
			p = binTempFileName + strlen(binTempFileName);
		strcpy(p, ".BI$");

		// BIN e|t@C쐬
		if (BINCreate(binTempFileName, &bfpTemp) != 0) {
			DeleteFile(dbfBackupFileName);
			free(rbp);
			return ERROR_DB_TMPFILE;		
		}

		// BIN t@CRs[
		for (r_rno = 0; r_rno < dhp->nRec; ) {
			if ((err = DBFRead(dp, ++r_rno, rbp)) != 0)
				break;

			bWriteBinary = FALSE;
			for (i = 0, flp = dp->flp; i < dp->nField; i++, flp++) {
				if (flp->type == 'B') {
					fdp = &rbp[flp->offset + 1];
					if (*(LONG *)fdp != 0) {
						if ((err = BINCopy(bfp, bfpTemp, *(LONG *)fdp, &bno)) != 0)
							goto end_binpack;
						*(LONG *)fdp = bno;
						bWriteBinary = TRUE;
					}
				}
			}

			if (bWriteBinary) {
				if ((err = DBFWrite(dp, r_rno, rbp)) != 0)
					break;
			}
		}
end_binpack:
		if (err == 0) {
			BINClose(bfpTemp);
			strcpy(binFileName, bfp->fileName);
			BINClose(bfp);
			DeleteFile(binFileName);
			MoveFile(binTempFileName, binFileName);
			BINOpen(binFileName, &dp->bfp);
			DeleteFile(dbfBackupFileName);
		} else {
			BINClose(bfpTemp);
			DeleteFile(binTempFileName);

			__CLOSE(dp->fh);
			DeleteFile(dp->fileName);
			MoveFile(dbfBackupFileName, dp->fileName);
			dp->fh = __OPEN(dp->fileName);
		}
	}	

	free(rbp);

	return 0;
}

/*=======================================================================
 |
 |		폜R[h`FbN
 |
 |	SHORT	DBFCheckDeleted(dp, rno, flag)
 |
 |		PDBF	dp;		cae|C^
 |		LONG	rno;		R[hԍ
 |		SHORT	*flag;		OFʏ탌R[h@PF폜R[h
 |
 |		Ԓl			G[R[h
 |
 =======================================================================*/
SHORT	DBFCheckDeleted(PDBF dp, LONG rno, SHORT *flag)
{
	PDBF_H	dhp;
	CHAR	buf[1];
#if defined (OS_WNT)
	DWORD	numBytes;
#endif

	dhp = dp->dhp;

	/* R[hԍ͈͊OȂG[ */
	if (rno <= 0 || rno > dhp->nRec)
		return ERROR_DB_RECNO;

	/* 폜tOZbg */
	__LSEEK(dp->fh, dhp->lHeader + dhp->lRec*(rno-1), SEEK_SET);
	if (__READ(dp->fh, buf, 1) != 1)
		return ERROR_DB_READ;

	*flag = (buf[0] == '*') ? 1 : 0;

	return 0;
}

/*=======================================================================
 |
 |		f[^t@CRs[
 |
 |	SHORT	DBFCopy(dp1, dp2)
 |
 |		PDBF	dp1;		Rs[cae|C^
 |		PDBF	dp2;		Rs[cae|C^
 |
 |		Ԓl			G[R[h
 |
 =======================================================================*/
SHORT	DBFCopy(PDBF dp1, PDBF dp2)
{
	LONG	rno;
	SHORT	i, j, len;
	PFIELD	flp1, flp2;
	CHAR	*rbp1, *rbp2;
	PBIN	bfp1, bfp2;
	LONG	bno;
	SHORT	err;

	if ((rbp1 = malloc(dp1->dhp->lRec)) == NULL)
		return ERROR_DB_MEMORY;

	if ((rbp2 = malloc(dp2->dhp->lRec)) == NULL) {
		free(rbp1);
		return ERROR_DB_MEMORY;
	}

	/* R[hZbg */
	dp2->dhp->nRec = dp1->dhp->nRec;

	bfp1 = dp1->bfp;
	bfp2 = dp2->bfp;

	err = 0;

	for (rno = 1; rno <= dp1->dhp->nRec; rno++) {
		if ((err = DBFRead(dp1, rno, rbp1)) != 0)
			break;

		memset(rbp2, ' ', dp2->dhp->lRec);
		for (i = 0; i < dp2->nField; i++) {
			flp2 = &dp2->flp[i];
			for (j = 0; j < dp1->nField; j++) {
				flp1 = &dp1->flp[j];
				if (strcmp(flp1->name, flp2->name) == 0)
					break;
			}
			if (j < dp1->nField) {
				len = min(flp1->lField2, flp2->lField2);
				switch (flp2->type) {
				case 'N':
				case 'I':
					memcpy(&rbp2[flp2->offset + (flp2->lField2-len) + 1],
							&rbp1[flp1->offset + (flp1->lField2-len) + 1], len);
					break;
				case 'B':
					if (bfp1 != NULL && bfp2 != NULL) {
						BINCopy(bfp1, bfp2, *(LONG *)&rbp1[flp1->offset + 1], &bno);
						*(LONG *)&rbp2[flp2->offset + 1] = bno;
					}
					break;
				default:
					memcpy(&rbp2[flp2->offset + 1], &rbp1[flp1->offset + 1], len);
					break;
				}
			}
		}
		if ((err = DBFWrite(dp2, rno, rbp2)) != 0)
			break;
	}

	free(rbp1);
	free(rbp2);

	if (err != 0)
		return err;

	/* t@CTCY𒲐 */
	__CHSIZE(dp2->fileName, dp2->fh,
			dp2->dhp->lHeader + dp2->dhp->lRec * (rno - 1));

	return 0;
}

/*=======================================================================
 |
 |		SẴR[h폜
 |
 |	SHORT	DBFZip(dp)
 |
 |		PDBF	dp;		cae|C^
 |
 |		Ԓl			G[R[h
 |
 =======================================================================*/
SHORT	DBFZip(PDBF dp)
{
	PDBF_H	dhp;

	dhp = dp->dhp;

	dhp->nRec = 0;		/* R[hOɂ */

	/* t@CTCY𒲐 */
	__CHSIZE(dp->fileName, dp->fh, (LONG)dhp->lHeader);

	/* oCif[^̑S폜 */
	if (dp->bfp != NULL)
		BINZip(dp->bfp);

	dp->updateFlag = 1;		/* XVtOZbg */
	dp->updateCount++;

	return 0;
}

/*=======================================================================
 |
 |		ÍL[ݒ
 |
 |	SHORT	DBFSetScramble(dp, scrambleKey)
 |
 |		PDBF	dp;		cae|C^
 |		USHORT	scramble;	ÍL[
 |
 |		Ԓl			G[R[h
 |
 =======================================================================*/
SHORT	DBFSetScramble(PDBF dp, USHORT scramble)
{
	PDBF_H	dhp;
	USHORT	currentScramble;
	LONG	recNo;
	CHAR	*rbp;
	SHORT	err;

	dhp = dp->dhp;

	if ((rbp = malloc(dhp->lRec)) == NULL)
		return ERROR_DB_MEMORY;

	currentScramble = dhp->scramble;
	for (recNo = 1; recNo <= dhp->nRec; recNo++) {
		dhp->scramble = currentScramble;
		if ((err = DBFRead(dp, recNo, rbp)) != 0) {
			free(rbp);
			return err;
		}
		dhp->scramble = scramble;
		if ((err = DBFWrite(dp, recNo, rbp)) != 0) {
			free(rbp);
			return err;
		}
	}
	dhp->scramble = scramble;

	free(rbp);

	return 0;
}

/*=======================================================================
 |
 |		pX[hݒ
 |
 |	SHORT	DBFSetPassword(dp, password)
 |
 |		PDBF	dp;		cae|C^
 |		CHAR	*password;	pX[h
 |
 |		Ԓl			G[R[h
 |
 =======================================================================*/
SHORT	DBFSetPassword(PDBF dp, CHAR *password)
{
	PDBF_H	dhp;

	dhp = dp->dhp;

	if (password == NULL || password[0] == '\0') {
		memset(dhp->password, 0, 8);
	} else {
		memset(dhp->password, ' ', 8);
		memcpy(dhp->password, password, min(strlen(password), 8));
		Scramble1(dhp->password + 1, 7, dhp->password[0]);
	}

	dp->updateFlag = 1;

	return 0;
}
